# -*- encoding: utf-8

import os
import shutil

import bs4
import htmlmin

from . import css, templates
from .settings import create_new_settings

ROOT = os.path.dirname(os.path.abspath(__file__))

OUTPUT_DIR = '_output'


def create_site(site_dir):
    """Perform the setup for a new site."""
    create_new_settings(site_dir)

    for dirname in ('pages', 'posts', 'static'):
        os.makedirs(os.path.join(site_dir, dirname), exist_ok=True)

    for dirname in [css.STYLE_DIR, templates.TEMPLATE_DIR]:
        shutil.copytree(
            os.path.join(ROOT, dirname),
            os.path.join(site_dir, dirname)
        )


def clean_site(site_dir):
    """Delete all the generated output."""
    try:
        shutil.rmtree(OUTPUT_DIR)
    except FileNotFoundError:
        pass


class NewSite:
    """
    Mixin class containing refactored methods from the ``Site`` class
    in ``__init__.py``.
    """
    def add_css_to_html(self, css_string):
        """Insert CSS into all the rendered HTML pages."""
        for slug, html_str in self._prepared_html.items():

            # Find the body text for this page -- i.e., the HTML we want
            # to compare the CSS with.
            soup = bs4.BeautifulSoup(html_str, 'html.parser')
            body_html = ''.join([str(s) for s in soup.find('body').contents])

            min_css = css.minimal_css_for_html(
                body_html=body_html, css=css_string
            )

            # Add the CSS within the <head> section.  If there isn't one,
            # add it.
            if min_css.strip():
                if '</head>' not in html_str:
                    html_str = html_str.replace('<body', '<head></head><body', 1)
                html_str = html_str.replace(
                    '</head>', f'<style>{min_css}</style></head>'
                )

            self._prepared_html[slug] = html_str

    def optimise_html(self):
        """Minify the HTML and prepare it for writing to disk."""
        for slug, html_str in self._prepared_html.items():
            html_str = htmlmin.minify(html_str)
            self._prepared_html[slug] = html_str



# -*- encoding: utf-8
"""
Implements the Markdown parser.

This module exports a single function, which converts Markdown source into
HTML, and then returns the HTML.  It doesn't do any I/O -- everything is just
pure in-memory strings.
"""

import markdown
from markdown.extensions.codehilite import CodeHiliteExtension
from markdown.extensions.footnotes import FootnoteExtension
from markdown.extensions.fenced_code import FencedCodeExtension
from markdown.extensions.meta import MetaExtension
from markdown.extensions.smarty import SmartyExtension


def convert_markdown(source, extra_extensions=None, parse_metadata=True):
    """
    Convert a Markdown string into HTML.  Returns an (HTML, metadata) tuple.

    This renders vanilla Markdown, and also footnotes, fenced code blocks
    and syntax-highlighting for code blocks.  Additional extensions can
    be supplied in the ``extra_extensions`` parameter.

    :param source: Markdown source.
    :param extra_extensions: (optional) A list of Extensions to use.
    :param parse_metadata: (optional) Should we parse metadata in the string?

    """
    # Rather than load all the extensions every time, we load only the
    # extensions required for the given source.  Informal testing in the old
    # implementation suggested this made a difference.
    # TODO: Is this still required?
    extensions = [SmartyExtension()]

    if parse_metadata:
        extensions.append(MetaExtension())

    # Look for evidence of footnotes.
    if '[^' in source:
        fn_extension = FootnoteExtension(
            # We may show multiple documents with footnotes on an index
            # page.  Ensure footnote references are globally unique.
            # TODO: Make footnote numbering consistent over multiple
            # builds even when the set of pages/posts changes.
            UNIQUE_IDS=True,

            # Make sure that footnote markers are rendered as a text
            # arrow on iOS devices, not emoji.  For more info:
            # http://daringfireball.net/linked/2015/04/22/unicode-emoji
            BACKLINK_TEXT='&#8617;&#xFE0E;',
        )
        extensions.append(fn_extension)

    if '```' in source:
        extensions.extend([
            CodeHiliteExtension(),
            FencedCodeExtension(),
        ])

    if extra_extensions is not None:
        extensions.extend(extra_extensions)

    md = markdown.Markdown(extensions=extensions)
    html = md.convert(source)

    if parse_metadata:
        return (html, md.Meta)
    else:
        return html



# -*- encoding: utf-8
"""
Provides template utilities.
"""

import functools

import jinja2

from . import markdown as md, plugins

# TODO: Make this a setting
TEMPLATE_DIR = 'templates'


# TODO: Make this a setting
def locale_date(date):
    """Render a date in the current locale date."""
    return date.strftime('%d %B %Y').lstrip('0')


@functools.lru_cache()
def render_title(title):
    return md.convert_markdown(
        title,
        extra_extensions=plugins.load_markdown_extensions(),
        parse_metadata=False
    ).replace('<p>', '').replace('</p>', '')


def build_environment(template_dir=None):
    """Build the appropriate Jinja2 environment."""
    if template_dir is None:
        template_dir = TEMPLATE_DIR
    env = jinja2.Environment(
        loader=jinja2.FileSystemLoader(template_dir),
        autoescape=jinja2.select_autoescape(['html']),
        undefined=jinja2.StrictUndefined
    )

    # TODO: Extension mechanism for additional filters?
    env.filters['locale_date'] = locale_date
    env.filters['title'] = render_title

    return env



# -*- encoding: utf-8


class bcolors:
    info = '\033[0;34m'
    success = '\033[0;32m'
    ENDC = '\033[0m'


def _print(color, message, args):
    msg = message % args
    print('[' + getattr(bcolors, color) + color + bcolors.ENDC + '] ' + msg)


def info(message, *args):
    _print(color='info', message=message, args=args)


def success(message, *args):
    _print(color='success', message=message, args=args)



# -*- encoding: utf-8

# flake8: noqa

import collections
import concurrent.futures
from datetime import date, datetime
import itertools
import os
import re
import sys

import dateutil.parser as dp
import htmlmin
from feedgenerator import Atom1Feed, get_tag_uri

from . import logging, markdown, plugins
from .css import load_base_css, minimal_css_for_html, optimize_css
from .logging import info
from .readers import list_page_files, list_post_files
from .site import NewSite
from .settings import load_settings, validate_settings
from .templates import build_environment
from .utils import Pagination, lazy_copyfile, slugify


# TODO: Make this a setting
PAGE_SIZE = 10


class Site(NewSite):
    """Represents a single site."""

    def __init__(self, settings):
        self.path = os.path.abspath(os.curdir)
        self.settings = settings
        validate_settings(self.settings)

        self.posts = []
        self.pages = []
        self.env = build_environment()

        # Mapping from URL slugs to rendered HTML.
        # TODO: Check we don't write the same slug more than once.
        self._prepared_html = {}

    def write(self):
        """Write all the prepared HTML to files on disk."""
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            futures = {
                executor.submit(self.write_html, slug, html_str): slug
                for slug, html_str in self._prepared_html.items()
            }
            for future in concurrent.futures.as_completed(futures):
                logging.info('Written HTML for %s', futures[future])

    def write_html(self, slug, html_str):
        """
        Writes a string to a path in the output directory.

        This creates a directory with an ``index.html`` file, so you get
        pretty URLs without needing web server configuration.
        """
        slug = slug.lstrip('/')
        os.makedirs(os.path.join(site.OUTPUT_DIR, slug), exist_ok=True)

        with open(os.path.join(site.OUTPUT_DIR, slug, 'index.html'), 'w') as f:
            f.write(html_str)

    def build(self):
        """
        Build the complete site and write it to the output folder.
        """
        self.posts = sorted(self.posts, key=lambda x: x.metadata['date'], reverse=True)

        self._prepare_posts()
        self._prepare_pages()
        self._prepare_index(posts=self.posts)
        self._prepare_date_index()
        self._prepare_archive()
        self._prepare_tag_index()

        css_string = optimize_css(load_base_css())
        self.add_css_to_html(css_string=css_string)

        self.optimise_html()

        self._build_feeds()
        self._copy_static_files()
        self.write()

    def _prepare_posts(self):
        """Prepare the HTML for posts."""
        post_template = self.env.get_template('post.html')
        for post in self.posts:
            html = post_template.render(
                site=self,
                metadata=post.metadata,
                content=post.content
            )
            self._prepared_html[post.metadata['url']] = html

    def _prepare_pages(self):
        """Prepare the HTML for pages."""
        page_template = self.env.get_template('page.html')
        for page in self.pages:
            html = page_template.render(
                site=self,
                metadata=page.metadata,
                content=page.content
            )
            self._prepared_html[page.metadata['url']] = html

    def _prepare_index(self, posts, prefix='', title=None):
        """Prepare the HTML for a set of index pages."""
        index_template = self.env.get_template('index.html')
        posts = sorted(posts, key=lambda x: x.metadata['date'], reverse=True)

        pagination = Pagination(
            posts=posts, page_size=PAGE_SIZE, prefix=prefix
        )

        for pageset in pagination:
            html = index_template.render(
                site=self,
                posts=pageset.posts,
                title=title,
                pageset=pageset
            )
            self._prepared_html[pageset.slug] = html

    def _prepare_date_index(self):
        """Prepare the HTML for the date-based index."""
        def get_month(p):
            return date(p.metadata['date'].year, p.metadata['date'].month, 1)

        for m, posts in itertools.groupby(self.posts, get_month):
            self._prepare_index(
                posts=posts,
                prefix='/%04d/%02d' % (m.year, m.month),
                title=m.strftime('Posts from %B %Y')
            )

    def _prepare_archive(self):
        archive_template = self.env.get_template('archive.html')

        def get_year(p):
            return p.metadata['date'].year

        html = archive_template.render(
            site=self,
            post_groups=itertools.groupby(self.posts, get_year),
            title='Archives'
        )
        self._prepared_html['/archives/'] = html

    def _prepare_tag_index(self):
        """Prepare the HTML for the tag-based index."""
        tags = collections.defaultdict(list)
        for p in self.posts:
            for t in p.metadata['tags']:
                tags[t].append(p)

        for t, posts in tags.items():
            self._prepare_index(
                posts=posts,
                prefix='/tag/%s' % t,
                title='Tagged with “%s”' % t
            )

    @classmethod
    def from_folder(cls, path):
        """Construct a ``Site`` instance from a folder on disk."""
        site = cls(settings=load_settings(path))
        for path in list_post_files(site.path):
            info('Reading post from file %s',
                path.replace(site.path, '').lstrip('/'))
            site.posts.append(Post.from_file(path))

        for path in list_page_files(site.path):
            info(
                'Reading page from file %s',
                path.replace(site.path, '').lstrip('/'))
            site.pages.append(Page.from_file(path))

        return site

    def _build_feeds(self):
        feed_kwargs = {
            'title': self.settings['name'],
            'link': self.settings['url'] + '/feeds/all.atom.xml',
            'description': self.settings['description'],
            'author_name': self.settings['author'],
            'author_email': self.settings['author_email'],
        }
        feed = Atom1Feed(**feed_kwargs)

        for post in reversed(self.posts):
            post_kwargs = {
                'title': post.metadata['title'],
                'link': self.settings['url'] + post.metadata['url'],
                'pubdate': post.metadata['date'],
                'content': post.content,
            }
            if post.metadata.get('link') is not None:
                post_kwargs['link'] = post.metadata.get('link')
            if 'summary' in post.metadata:
                post_kwargs['description'] = post.metadata['summary']
            else:
                post_kwargs['description'] = post_kwargs['content']
            post_kwargs['unique_id'] = get_tag_uri(post_kwargs['link'], post.metadata['date'])

            if '<blockquote class="update">' in post.content:
                update_strings = re.findall(
                    r'Update, [0-9]+ [A-Z][a-z]+ [0-9]{4}', post.content
                )
                updated = max([
                    datetime.strptime(u, 'Update, %d %B %Y')
                    for u in update_strings
                ])
                post_kwargs['updateddate'] = updated

            feed.add_item(**post_kwargs)

        feed_dir = os.path.join(site.OUTPUT_DIR, 'feeds')
        os.makedirs(feed_dir, exist_ok=True)
        feed.write(
            open(os.path.join(feed_dir, 'all.atom.xml'), 'w'),
            encoding='utf-8')

    def _copy_static_files(self):
        for root, _, filenames in os.walk(os.path.join(self.path, 'static')):
            for f in filenames:
                if f.startswith('.'):
                    continue
                base = os.path.join(root, f).replace(
                    self.path + '/static/', '')
                lazy_copyfile(
                    src=os.path.join(self.path, 'static', base),
                    dst=os.path.join(site.OUTPUT_DIR, base),
                )


class Article:
    """
    Holds information about an individual article (a page or a post).
    """
    def __init__(self, content, metadata, path):
        self.content = content
        self.metadata = metadata
        self.path = path

        if 'slug' not in self.metadata:
            self.metadata['slug'] = slugify(self.metadata['title'])

        self.metadata['url'] = self.metadata['slug']

        try:
            self.metadata['tags'] = sorted([
                t.strip()
                for t in self.metadata.get('tags', '').split(',')
                if t.strip()])
        except KeyError:
            pass

        self.metadata = metadata

    def __repr__(self):
        return '<%s path=%r>' % (type(self).__name__, self.path)

    @classmethod
    def from_string(cls, path, file_contents):
        """Construct an instance from a string read from a file."""
        html, metadata = markdown.convert_markdown(
            file_contents,
            extra_extensions=plugins.load_markdown_extensions()
        )

        metadata = {k: v[0] for k, v in metadata.items()}

        return cls(
            content=html,
            metadata=metadata,
            path=path)

    @classmethod
    def from_file(cls, path):
        """Construct an ``Article`` instance from a file on disk."""
        file_contents = open(path).read()
        return cls.from_string(path=path, file_contents=file_contents)


# Add some error checking that these have the correct fields

class Page(Article):
    """Holds information about an individual page."""
    pass


class Post(Article):
    """Holds information about an individual post."""
    def __init__(self, content, metadata, path):
        super().__init__(content, metadata, path)
        self.metadata['date'] = dp.parse(self.metadata['date'])
        self.metadata['url'] = self.metadata['date'].strftime('%Y/%m/') + self.metadata['slug']



# -*- encoding: utf-8

import collections
import filecmp
import os
import re
import shutil

from unidecode import unidecode


def slugify(u):
    """Convert Unicode string into blog slug."""
    # http://www.leancrew.com/all-this/2014/10/asciifying/
    u = re.sub(u'[–—/:;,.]', '-', u)   # replace separating punctuation
    a = unidecode(u).lower()           # best ASCII substitutions, lowercased
    a = re.sub(r'[^a-z0-9 -]', '', a)  # delete any other characters
    a = a.replace(' ', '-')            # spaces to hyphens
    a = re.sub(r'-+', '-', a)          # condense repeated hyphens
    return a


def chunks(l, n):
    """Yield successive n-sized chunks from l."""
    # http://stackoverflow.com/a/312464/1558022
    # Note: this only works if we know hte length of ``l``.
    for i in range(0, len(l), n):
        yield l[i:i + n]


def lazy_copyfile(src, dst):
    """
    Copy a file from ``src`` to ``dst``, but only if the files are different.

    Avoids thrashing the disk unnecessarily on repeated site builds.
    """
    assert os.path.exists(src)
    if os.path.exists(dst):
        if os.path.getmtime(dst) >= os.path.getmtime(src):
            return
        elif filecmp.cmp(src, dst):
            return
    os.makedirs(os.path.dirname(dst), exist_ok=True)
    shutil.copyfile(src, dst)


Pageset = collections.namedtuple('Pageset', ['slug', 'posts', 'next', 'prev'])


class Pagination:
    """
    Provides pagination for a series of posts.

    Given a series of posts, the page size and the URL prefix
    (i.e. what's the URL slug for this series of posts), you can iterate
    over this object to get the collection of posts, slug and next/prev links
    for these posts.
    """
    def __init__(self, posts, page_size, prefix):
        self.posts = posts
        self.size = page_size
        self.prefix = prefix.strip('/')

    def __iter__(self):
        for pageno, posts in enumerate(chunks(self.posts, self.size), start=1):
            if pageno == 1:
                if self.prefix:
                    slug = '/%s/' % self.prefix
                else:
                    slug = '/'
                prev_ = None
            else:
                prefix = self.prefix or 'page'
                prev_ = slug
                slug = '/%s/%d/' % (prefix, pageno)

            if self.size * pageno >= len(self.posts):
                next_ = None
            else:
                prefix = self.prefix or 'page'
                next_ = '/%s/%d/' % (prefix, pageno + 1)

            yield Pageset(slug, posts, next_, prev_)



# -*- encoding: utf-8

import os

MARKDOWN_EXTENSIONS = ('.txt', '.md', '.mdown', '.markdown')


def read_isolate_file(site_dir):
    try:
        with open(os.path.join(site_dir, '.isolate')) as f:
            name = f.read().strip()
        return name or None
    except FileNotFoundError:
        pass


def _get_files(site_dir, subdir):
    isolated = read_isolate_file(site_dir)
    if (isolated is not None) and (subdir not in isolated):
        return

    for root, _, filenames in os.walk(os.path.join(site_dir, subdir)):
        for f in filenames:
            if f.lower().endswith(MARKDOWN_EXTENSIONS):
                if (isolated is None) or (isolated.endswith(f)):
                    yield os.path.join(root, f)


def list_post_files(site_dir):
    yield from _get_files(site_dir, subdir='posts')


def list_page_files(site_dir):
    yield from _get_files(site_dir, subdir='pages')



# -*- encoding: utf-8

import glob
import os
import sys

import markdown


class MarkdownExtensionMixin(markdown.Extension):
    """Stub class for importing additional Markdown extensions."""
    pass


def load_plugins(plugin_dir):
    """Load any plugins from the plugin directory."""
    orig_path = list(sys.path)
    sys.path.append(os.path.basename(plugin_dir))
    for path in glob.iglob('%s/*.py' % plugin_dir):
        __import__(os.path.basename(path).replace('.py', ''))
    sys.path = orig_path


def load_markdown_extensions():
    """Return any ``MarkdownExtension`` plugins."""
    load_plugins(os.path.join(os.path.abspath(os.curdir), 'plugins'))
    return [e() for e in MarkdownExtensionMixin.__subclasses__()]



# -*- encoding: utf-8

import os
import shutil
import sys
import time

import click
import docker
import requests

from . import site, Site
from .logging import success


@click.group()
def cli():
    pass


@cli.command('init', help='create a new site')
def init():
    site.create_site(os.curdir)


@cli.command('build', help='(re)build the HTML for the website')
def build():
    start = time.time()
    site = Site.from_folder(os.curdir)
    site.build()
    success('Site build completed in %.2f seconds', time.time() - start)


@cli.command('clean', help='remove the generated HTML')
def clean():
    # TODO: This will clobber the Docker container's volume mount, if present
    # Stop and delete the container if it's running.
    site.clean_site(os.curdir)


@cli.command('serve', help='serve the generated website on a local port')
@click.option('--port', default=9000, help='port for the web server',
              type=click.IntRange(1024, 65535))
def serve(port):
    """
    Start a local web server for the site.
    """
    site = Site.from_folder('content')
    site.build()

    # Set up the Docker API
    client = docker.from_env()

    # Check that Docker is available
    try:
        client.info()
    except requests.exceptions.ConnectionError:
        sys.exit('Unable to connect to Docker.')

    # Start the container
    try:
        client.containers.run(
            image='httpd',
            detach=True,
            ports={80: port},
            volumes={
                os.path.abspath(site.out_path): {
                    'bind': '/usr/local/apache2/htdocs',
                    'mode': 'ro',
                },
            },
        )
    except docker.errors.APIError as err:
        sys.exit('Error starting the Docker container:\n%s' % err)

    print('Web server running on http://localhost:%d/' % port)


@cli.command('publish', help='build the HTML for publication')
def publish():
    """
    The ``build`` command won't clean up pages or links that no longer exist.
    This command builds a fresh copy of the site, free of dead links.
    """
    for f in os.listdir('output'):
        path = os.path.join('output', f)
        if os.path.isdir(path):
            shutil.rmtree(path)
        else:
            os.unlink(path)

    site = Site.from_folder('content')
    site.build()



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import codecs
import os

from setuptools import find_packages, setup


def local_file(name):
    return os.path.relpath(os.path.join(os.path.dirname(__file__), name))


SOURCE = local_file('src')
README = local_file('README.rst')
long_description = codecs.open(README, encoding='utf-8').read()


setup(
    name='hotchocolate',
    version='3.0.1',
    description='A static site generator designed for lightweight sites',
    long_description=long_description,
    url='https://github.com/alexwlchan/hot-chocolate',
    author='Alex Chan',
    author_email='alex@alexwlchan.net',
    license='MIT',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Other Audience',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3 :: Only',
    ],
    packages=find_packages(SOURCE),
    package_dir={'': SOURCE},
    include_package_data=True,
    install_requires=[
        'beautifulsoup4>=4.6.0,<5',
        'click>=6.7,<7',
        'csscompressor>=0.9.4,<1',
        'docker>=2.0.2,<3',
        'feedgenerator>=1.9,<2',
        'htmlmin>=0.1.10,<0.2',
        'python-dateutil>=2.6.0,<3',
        'Jinja2>=2.9.5,<3',
        'Markdown>=2.6.8,<3',
        'mincss>=0.11.2,<0.12',
        'Pygments>=2.2.0,<3',
        'pyScss>=1.3.5,<2',
        'toml>=0.9.2,<1',
        'unidecode>=0.04.20,<0.05',
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'cocoa=hotchocolate.cli:cli',
        ],
    },
)



# -*- encoding: utf-8
"""
Implements the CSS processors.

Because static site generators do all their processing upfront, you can
spend time in the build step optimising the CSS for the smallest size.
Remove duplicate selectors, optimise layout, minify, etc.

Hot Chocolate goes one step further -- it's designed for sites with a very
small CSS footprint (single-digit KB), so it puts the CSS in a <style> tag
within the <head> of the page -- after reducing it to only the selectors
required to render the current page.  This means that every page loads the
smallest amount of CSS it can get away with.

"""

import os
import subprocess
import tempfile

import csscompressor
import mincss.processor as mp
import scss.compiler

# TODO: Make this a setting
STYLE_DIR = 'styles'


def optimize_css(css):
    """Given a CSS string, optimize and minify it as much as possible."""
    # Run it through the cleancss tool
    css = cleancss(css)

    # Minify it, stripping out comments and whitespace.  This is arguably
    # redundant because cleancss can do this for us, but it's a good fallback
    # if cleancss fails for some reason.
    css = csscompressor.compress(css)

    return css


def cleancss(css):
    """
    Run the CSS through the cleancss Node tool
    (https://github.com/jakubpawlowicz/clean-css).

    If the tool is unavailable, it just returns the CSS string unmodified.
    """
    # To run cleancss, the CSS must be in a file.
    _, path = tempfile.mkstemp()
    with open(path, 'w') as fp:
        fp.write(css)

    try:
        return subprocess.check_output(
            ['cleancss', '-O', '2', path]).decode('utf8')
    except (OSError, subprocess.CalledProcessError):
        return css
    finally:
        os.unlink(path)


class _InMemoryProcessor(mp.Processor):
    """A wrapper for doing in-memory stripping of unused CSS rules."""
    # The internals of this class are a bit messy and don't present themselves
    # well to doing in-memory operations: they expect to go out to the
    # filesystem and read files.  We have to change a few internal pieces
    # as we already have the file to hand.

    def download(self, url):
        return url

    def find_covering_css(self, body_html, css):
        self.process(f'<style>{css}</style><body>{body_html}</body>')
        try:
            return self.inlines[0].after
        except IndexError:
            return ''


def minimal_css_for_html(body_html, css):
    """Returns the minimal CSS required to render a block of body HTML."""
    proc = _InMemoryProcessor()
    return proc.find_covering_css(body_html=body_html, css=css)


def load_base_css(style_dir=None):
    """
    Load the ``style.scss`` file from the site's ``style`` directory, and
    return the base CSS for the site.
    """
    if style_dir is None:
        style_dir = STYLE_DIR

    return scss.compiler.compile_file(os.path.join(style_dir, 'style.scss'))



# -*- encoding: utf-8

import os
import shutil

import toml


SETTINGS_NAME = 'settings.toml'

REQUIRED_SETTINGS = {
    'name',
    'url',
    'author',
    'author_email',
    'description',
    'language',
}


def _settings_path(site_dir):
    return os.path.join(site_dir, SETTINGS_NAME)


class UndefinedSetting(Exception):
    pass


def validate_settings(settings):
    keys = set(settings.keys())
    missing_keys = REQUIRED_SETTINGS - keys
    if missing_keys:
        raise UndefinedSetting(
            f"Missing required settings: {', '.join(missing_keys)}"
        )


def create_new_settings(site_dir):
    """Create a new settings directory in the given directory."""
    shutil.copyfile(
        os.path.join(_settings_path(os.path.dirname(os.path.abspath(__file__)))),
        os.path.join(site_dir, 'settings.toml')
    )


def load_settings(site_dir):
    """Loads the settings from a given directory."""
    path = _settings_path(site_dir)
    return toml.loads(open(path).read())['hotchocolate']
